/*
 * ilbctool - command line utility to encode/decode iLBC
 * Copyright (C) 2008  Tommi Saviranta  <wnd@iki.fi>
 *
 * This application is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * Note that this application uses and statically links to example
 * implementation of iLBC, internet Low Bitrate Codec. Example code
 * for iLBC is released under GIPS licence, available at
 * http://ilbcfreeware.org/documentation/gips_iLBClicense.pdf .
 * This package should not include the RFC or the sources but it does
 * download the former and separate the latter from it.
 *
 * This application is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/* FIXME: LE handling in writing header */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "iLBC_define.h"
#include "iLBC_encode.h"
#include "iLBC_decode.h"

#include <assert.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>

#define ILBCNOOFWORDS_MAX	(NO_OF_BYTES_30MS / 2)
#define RIFF_HEADER		44


enum type {
	TYPE_UNSET = 0,
	TYPE_RAW,
	TYPE_RIFF,
	TYPE_ILBC,
	TYPE_UNKNOWN
};



static char *
make_filename(const char *src, const char *suffix)
{
	const char *ptr;
	char *new;
	int chop;

	ptr = strrchr(src, (int) '.');
	if (! ptr) {
		chop = strlen(src);
	} else {
		chop = (int) (ptr - src);
	}

	new = malloc(chop + strlen(suffix) + 1);
	if (! new) {
		return NULL;
	}
	memcpy(new, src, chop);
	sprintf(new + chop, "%s", suffix);
	return new;
}



static uint32_t
encode(FILE *in, FILE *out, short mode)
{
	short encoded_data[ILBCNOOFWORDS_MAX], decoded_data[BLOCKL_MAX];
	float block[BLOCKL_MAX];
	iLBC_Enc_Inst_t encoder;

	int i;
	short len;
	uint32_t size = 0;

	initEncode(&encoder, mode);

	while (1) {
		len = (short) fread(decoded_data, sizeof(short),
				encoder.blockl, in);
		if (len == 0) {
			break;
		}

		for (i = 0; i < (int) encoder.blockl; i++) {
			block[i] = (float) decoded_data[i];
		}

		iLBC_encode((unsigned char *) encoded_data,
				block, &encoder);
		for (i = 0; i < (int) encoder.no_of_bytes; i++) {
			fputc(((char *) encoded_data)[i], out);
		}
		size += encoder.no_of_bytes;
	}

	return size;
}



static uint32_t
decode(FILE *in, FILE *out, short mode)
{
	short decoded_data[BLOCKL_MAX], encoded_data[ILBCNOOFWORDS_MAX];
	float block[BLOCKL_MAX];
	iLBC_Dec_Inst_t decoder;

	float t;
	int i;
	short len;
	uint32_t size = 0;

	initDecode(&decoder, mode, 1);

	while (1) {
		len = (short) fread(encoded_data, sizeof(unsigned char),
				decoder.no_of_bytes, in);
		if (len == 0) {
			break;
		}

		iLBC_decode(block, (unsigned char *) encoded_data,
				&decoder, 1);
		for (i = 0; i < (int) decoder.blockl; i++) {
			t = block[i];
			if (t < MIN_SAMPLE) {
				t = MIN_SAMPLE;
			} else if (t > MAX_SAMPLE) {
				t = MAX_SAMPLE;
			}
			decoded_data[i] = (short) t;
		}

		for (i = 0; i < (int) decoder.blockl * 2; i++) {
			fputc(((char *) decoded_data)[i], out);
		}
		size += decoder.blockl * 2;

	}

	return size;
}



#if !defined(WORDS_BIGENDIAN)
#define h16tole(a) (a)
#define h32tole(a) (a)
#else
static uint16_t
h16tole(uint16_t v)
{
	uint8_t *b = (uint8_t *) &v;
	uint8_t t;

	t = b[0];
	b[0] = b[1];
	b[1] = t;

	return *((uint16_t *) b);
}



static uint32_t
h32tole(uint32_t v)
{
	uint8_t *b = (uint8_t *) &v;
	uint8_t t;

	t = b[0];
	b[0] = b[3];
	b[3] = t;

	t = b[1];
	b[1] = b[2];
	b[2] = t;

	return *((uint32_t *) b);
}
#endif



static int
write_riff_16(FILE *out, uint16_t v)
{
	uint16_t t16 = h16tole(v);
	uint8_t *t = (uint8_t *) &t16;

	putc(t[0], out);
	putc(t[1], out);

	return 0;
}



static int
write_riff_32(FILE *out, uint32_t v)
{
	uint32_t t32 = h32tole(v);
	uint8_t *t = (uint8_t *) &t32;

	putc(t[0], out);
	putc(t[1], out);
	putc(t[2], out);
	putc(t[3], out);

	return 0;
}



static int
write_riff_header(FILE *out, uint32_t size)
{
	/* RIFF type chunk */

	fputc(0x52, out); 			/* 0x00: R */
	fputc(0x49, out); 			/* 0x01: I */
	fputc(0x46, out); 			/* 0x02: F */
	fputc(0x46, out); 			/* 0x03: F */

	write_riff_32(out, size + RIFF_HEADER);	/* 0x04: size */

	fputc(0x57, out); 			/* 0x08: W */
	fputc(0x41, out); 			/* 0x09: A */
	fputc(0x56, out); 			/* 0x0a: V */
	fputc(0x45, out); 			/* 0x0b: E */

	/* File chunk */

	/* "fmt " */
	fputc(0x66, out); 			/* 0x0c: f */
	fputc(0x6d, out); 			/* 0x0d: m */
	fputc(0x74, out); 			/* 0x0e: t */
	fputc(0x20, out); 			/* 0x0f:   */

	write_riff_32(out, 16);			/* 0x10: chunk data size */
	write_riff_16(out, 1);			/* 0x14: PCM/uncompressed */
	write_riff_16(out, 1);			/* 0x16: one channel */
	write_riff_32(out, 8000);		/* 0x18: sample rate */
	write_riff_32(out, 16000);		/* 0x1c: avg bytes per second */
	write_riff_16(out, 2);			/* 0x20: bytes per smpl slice */
	write_riff_16(out, 16);			/* 0x22: bits per sample */

	/* Data chunk */

	fputc(0x64, out);			/* 0x24: d */
	fputc(0x61, out);			/* 0x25: a */
	fputc(0x74, out);			/* 0x26: t */
	fputc(0x61, out);			/* 0x27: a */

	write_riff_32(out, size);		/* 0x28: data size */

	return 0;
}



static void
usage(void)
{
	printf("\
Usage: ilbctool [OPTION]... [FILE]...\n\
Encode or decode FILEs.\n\
\n\
  -c        write on standard output\n\
  -d        decompress\n\
  -h        give this help\n\
  -m=LEN    iLBC frame length in ms, default 30 (valid values: 20, 30)\n\
  -t=TYPE   use output type TYPE, default riff (valid types: raw, wav, riff)\n\
  -u        remove FILE after compression/decompression\n\
  -v        verbose\n\
  -V        display version number\n\
\n\
With no FILE, or when FILE in -, read standard output.\n\
");
}



static void
version(void)
{
	printf("%s %s\n", PACKAGE_NAME, PACKAGE_VERSION);
}



static enum type
get_type(const char *type)
{
	if (! strcasecmp(type, "wav") || ! strcasecmp(type, "riff")) {
		return TYPE_RIFF;
	} else if (! strcasecmp(type, "raw")) {
		return TYPE_RAW;
	} else {
		return TYPE_UNKNOWN;
	}
}



static const char *
get_type_suffix(enum type type)
{
	switch (type) {
		case TYPE_RAW:
			return ".raw";
		case TYPE_RIFF:
			return ".wav";
		case TYPE_ILBC:
			return ".ilbc";
		default:
			return "";
	}
}



int
main(int argc, char **argv)
{
	short ilbcmode = 30;
	int wout = 0;			/* write to stdout */
	int rm_after = 0;		/* unlink source file */
	int compress = 1;		/* compress */
	enum type type = TYPE_UNSET;	/* output type */
	int verbose = 0;

	while (1) {
		int opt;
		opt = getopt(argc, argv, "cdhm:t:uvV");
		if (opt == -1) {
			break;
		}

		switch (opt) {
			case 'c':
				wout = 1;
				if (type == TYPE_UNSET) {
					type = TYPE_RAW;
				}
				break;
			case 'd':
				compress = 0;
				break;
			case 'h':
				usage();
				return EXIT_SUCCESS;
			case 'm':
				ilbcmode = atoi(optarg);
				break;
			case 't':
				type = get_type(optarg);
				break;
			case 'u':
				rm_after = 1;
				break;
			case 'v':
				verbose = 1;
				break;
			case 'V':
				version();
				return EXIT_SUCCESS;
			case '?':
				usage();
				return EXIT_FAILURE;
		}
	}

	switch (type) {
		case TYPE_UNSET:
			type = TYPE_RIFF;
			break;
		case TYPE_UNKNOWN:
			usage();
			return EXIT_FAILURE;
		default:
			break;
	}

	if (wout && type == TYPE_RIFF) {
		fprintf(stderr, "%s: cannot decode to RIFF to stdout\n",
				argv[0]);
		return EXIT_FAILURE;
	}

	if (optind == argc) {
		if (compress) {
			encode(stdin, stdout, ilbcmode);
		} else {
			decode(stdin, stdout, ilbcmode);
			fprintf(stderr, "decode\n");
		}
	} else {
		int i;

		if (compress) {
			type = TYPE_ILBC;
		}

		for (i = optind; i < argc; i++) {
			FILE *in, *out;
			uint32_t size;

			in = fopen(argv[i], "rb");
			if (! in) {
				fprintf(stderr, "%s: cannot open file: %s\n",
						argv[0], argv[i]);
				return EXIT_FAILURE;
			}
			if (wout) {
				out = stdout;
				if (verbose) {
					fprintf(stderr, "%s -> stdout\n",
							argv[i]);
				}

			} else {
				char *name = make_filename(argv[i],
						get_type_suffix(type));
				assert(name);
				out = fopen(name, "wb");
				if (! out) {
					fprintf(stderr, "%s: cannot write "
							"file: %s\n",
							argv[0], name);
					free(name);
					fclose(in);
					return EXIT_FAILURE;
				}
				if (verbose) {
					fprintf(stderr, "%s -> %s\n",
							argv[i], name);
				}
				free(name);

				if (type == TYPE_RIFF) {
					fseek(out, RIFF_HEADER, SEEK_SET);
				}
			}

			if (compress) {
				size = encode(in, out, ilbcmode);
			} else {
				size = decode(in, out, ilbcmode);
			}

			if (! compress) {
				fseek(out, 0, SEEK_SET);
				switch (type) {
					case TYPE_RIFF:
						write_riff_header(out, size);
						break;
					default:
						break;
				}
			}

			if (rm_after) {
				unlink(argv[i]);
			}
			if (! wout) {
				fclose(out);
			}
			fclose(in);
		}
	}

	return EXIT_SUCCESS;
}

